/*
 *     jsz.Plotter - a Swing-based Java package for plotting data
 * Copyright (C) 2000  Janek Schwarz <schwarz@informatik.hu-berlin.de>
 *
 *                    This software is dedicated to 
 *                             THE RAMONES
 *                       the best punk band ever.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: Plotter.java,v 1.9 2000/02/03 21:01:39 schwarz Exp $
 */
package jsz.Plotter;

import javax.swing.*;
import java.util.*;
import java.awt.*;

/**
 * An object of this class implements a plotter, which is used
 * for visualizing data. This plotter class has some neat
 * properties. It is possible to specify, how a plotter behaves,
 * when data is larger then the coordinate system (ignore it, so
 * this datum isn't drawn, expand x and/or y dimension, so
 * everything fits in, or scroll, so older values scroll to the
 * left). There are different modes for drawing the caption
 * (mark every tick, mark just the maxima and mark
 * maxima and zero). However, it is possible to specify a class,
 * which does caption, so arbitrary captions can be
 * programmed.This class respects the current look-and-feel and
 * reacts on l&f changes.
 *
 * @author Janek Schwarz
 * @version 1.0
 */
public class Plotter
  extends JPanel
  implements Observer
{
/**
 * collection of IDataSeries', which have to be plotted
 */
private Vector plots = new Vector();

/**
 * logical dimension of the graph: minimal x-value
 */
private double xmin = 0.0;

/**
 * logical dimension of the graph: maximal x-value
 */
private double xmax = 1.0;

/**
 * logical dimension of the graph: minimal y-value
 */
private double ymin = 0.0;

/**
 * logical dimension of the graph: maximal y-value
 */
private double ymax = 1.0;

/**
 * each xTicksDistance a tick is drawn on the axis
 */
private double xTicksDistance = 0.1;

/**
 * each yTicksDistance a tick is drawn axis
 */
private double yTicksDistance = 0.1;

/**
 * it possible to draw dotted lines instead of ticks. Some might say, it looks
 * better (i.e, me)..
 */ 
private boolean ticksAsDottedLines = true;

/**
 * color, in which axis, ticks and caption are drawn
 */
private Color axisColor = UIManager.getColor("Label.foreground");

/**
 * font used for caption
 */
private Font axisFont = 
  new Font(UIManager.getFont("Label.font").getName(), 
           UIManager.getFont("Label.font").getStyle(), 10);

/* 
 * how to mark the axises:
 *     NONE: just draw the ticks, but don't mark them
 *     MAXIMA: just mark the maxima at both axises
 *     MAXIMA_ZERO: same as MAXIMA, but zero position as well.
 *     ALL: mark every tick
 *     OBJECT: pass every tick value to an object and let this object
 *             mark the tick. Could be used to map values to words.
 */
/**
 * used, when signaling "ignore something"
 */
public final static int NONE        = 0;

/**
 * style for inscribing ticks: just mark the maxima at both axises 
 */
public final static int MAXIMA      = 1;

/**
 * style for inscribing ticks: same as MAXIMA, but zero position as well
 */
public final static int MAXIMA_ZERO = 2;

/**
 * style for inscribing ticks: mark each tick
 */
public final static int ALL         = 3;

/**
 * style for inscribing ticks: pass every tick value to an object
 * and let this object mark the tick. Could be used to map
 * values to words.
 */
public final static int OBJECT      = 4;

/**
 * style with which the x-axis is marked
 */
private int xTicksCaptionStyle = MAXIMA_ZERO;

/**
 * style with which the y-axis is marked
 */
private int yTicksCaptionStyle = MAXIMA_ZERO;

/**
 * if <code>xTicksCaptionStyle</code> equals to
 * <code>OBJECT</code> then this variable contains a reference to
 * the caption object
 */
private ITickCaption xTicksCaptionObject = 
  new MaximaZeroCaption(this, true, false);

/**
 * if <code>yTicksCaptionStyle</code> equals to
 * <code>OBJECT</code> then this variable contains a reference to
 * the caption object
 */
private ITickCaption yTicksCaptionObject = 
  new MaximaZeroCaption(this, false, true);


/**
 * adjust {x|y}min and {x|y}max so the values fit in the graph
 */
public final static int FIT    = 1;

/**
 * adjust xmin and xmax so older values are scrolled to the left
 */
public final static int SCROLL = 2;

/**
 * how to behave when a value is larger than xmax or smaller than
 * xmin. Accepted values: NONE, FIT, SCROLL.
 */
private int xAxisBehaviour = FIT;

/**
 * how to behave when a value is larger than ymax or smaller than
 * ymin. Accepted values: NONE, FIT.
 */
private int yAxisBehaviour = FIT;

/**
 * height this SWING component
 */
private int currentWidth  = this.getWidth();

/**
 * width this SWING component
 */
private int currentHeight = this.getHeight();

/*
 * Constructor
 */
public Plotter()
{
  setDoubleBuffered(true);
  setOpaque(false);
}

/*
 * Constructor
 */
public Plotter(double xmin, 
               double xmax, 
               double ymin, 
               double ymax,
               double xTicksDistance, 
               double yTicksDistance)
{
  this.xmin           = xmin;  
  this.xmax           = xmax;
  this.ymin           = ymin;  
  this.ymax           = ymax;
  this.xTicksDistance = xTicksDistance;
  this.yTicksDistance = yTicksDistance;

  setDoubleBuffered(true);
  setOpaque(false);
}

/**
 * Add a new collection of data and plot it. The plotter adds
 * itself to the list of observers of ds, so it is notified about
 * changes of ds (adding data, changing color, legend etc.).
 */
public Plotter addPlot(IDataSeries ds)
{
  plots.addElement(ds);
  ds.addObserver(this);

  repaint();
  return this;
}

/**
 * set logical x-dimension (minimum)
 */
public final void setXMin(double xmin)
{
  this.xmin = xmin;
  repaint();
}

/**
 * get logical x-dimension (minimum)
 */
public final double getXMin()
{
  return xmin;
}

/**
 * set logical x-dimension (maximum)
 */
public final void setXMax(double xmax)
{
  this.xmax = xmax;
  repaint();
}

/**
 * get logical x-dimension (maximum)
 */
public final double getXMax()
{
  return xmax;
}

/**
 * Set the distance between to ticks on x-axis
 */
public final void setXticksDistance(double xTicksDistance)
{
  this.xTicksDistance = xTicksDistance;
  repaint();
}

/**
 * get the distance between to ticks on x-axis
 */
public final double getXticksDistance()
{
  return xTicksDistance;
}

/**
 * set logical y-dimension (minimum)
 */
public final void setYMin(double ymin)
{
  this.ymin = ymin;
  repaint();
}

/**
 * get logical y-dimension (minimum)
 */
public final double getYMin()
{
  return ymin;
}

/**
 * set logical y-dimension (maximum)
 */
public final void setYMax(double ymax)
{
  this.ymax = ymax;
  repaint();
}

/**
 * get logical y-dimension (maximum)
 */
public final double getYMax()
{
  return ymax;
}

/**
 * Set the distance between to ticks on y-axis
 */
public final void setYticksDistance(double yTicksDistance)
{
  this.yTicksDistance = yTicksDistance;
  repaint();
}

/**
 * get the distance between to ticks on y-axis
 */
public final double getYticksDistance()
{
  return yTicksDistance;
}

/**
 * set, wether ticks are drawn as dotted lines spawning the
 * component's background or as short lines crossing an axis.
 *
 * @param ticksAsDottedLines true: draw dotted lines
 */
public final void setTicksAsDottedLines(boolean ticksAsDottedLines)
{
  this.ticksAsDottedLines = ticksAsDottedLines;
  repaint();
}

/**
 * get, wether ticks are drawn as dotted lines spawning the
 * component's background
 */
public final boolean getTicksAsDottedLines()
{
  return ticksAsDottedLines;
}

/**
 * set, how x axis is marked.
 *
 * @param style: NONE, ALL, MAXIMA, MAXIMA_ZERO, OBJECT
 */
public final void setXticksCaptionStyle(int style)
  throws IllegalArgumentException
{
  switch(style)
    {
    case NONE:
      break;
    case ALL:
      xTicksCaptionObject = new AllCaption(false);
      break;
    case MAXIMA:
      xTicksCaptionObject = new MaximaCaption(this, true);
      break;
    case MAXIMA_ZERO:
      xTicksCaptionObject = new MaximaZeroCaption(this, true, false);
      break;
    case OBJECT:
      // the object itself is set with setXticksCaptionObject()
      break;
    default:
      throw new IllegalArgumentException();      
    }
  
  xTicksCaptionStyle = style;

  repaint();
}

/**
 * get, how x axis is marked.
 */
public final int getXticksCaptionStyle()
{
  return yTicksCaptionStyle;
}

/**
 * set, how y axis is marked.
 *
 * @param style: NONE, ALL, MAXIMA, MAXIMA_ZERO, OBJECT
 */
public final void setYticksCaptionStyle(int style)
  throws IllegalArgumentException
{
  switch(style)
    {
    case NONE:
      break;
    case ALL:
      yTicksCaptionObject = new AllCaption(true);
      break;
    case MAXIMA:
      yTicksCaptionObject = new MaximaCaption(this, false);
      break;
    case MAXIMA_ZERO:
      yTicksCaptionObject = new MaximaZeroCaption(this, false, true);
      break;
    case OBJECT:
      // the object itself is set with setYticksCaptionObject()
      break;
    default:
      throw new IllegalArgumentException();      
    }
  
  yTicksCaptionStyle = style;

  repaint();
}

/**
 * get, how y axis is marked.
 */
public final int getYticksCaptionStyle()
{
  return yTicksCaptionStyle;
}

/**
 * Mark the x-axis using an object. See the example provided
 * in {@link jsz.Plotter.ITickCaption}. This
 * method also changes <code>xTicksCaptionStyle</code> to
 * <code>OBJECT</code>.
 */
public final void setXticksCaptionObject(ITickCaption object)
{
  xTicksCaptionObject = object;
  xTicksCaptionStyle = OBJECT;
}

/**
 * return the caption object for the x-axis. Undefined
 * behaviour, if <code>xTicksCaptionStyle != OBJECT</code>.
 */
public final ITickCaption getXticksCaptionObject()
{
  return xTicksCaptionObject;
}

/**
 * Mark the y-axis using an object. See the example provided
 * in {@link jsz.Plotter.ITickCaption}. This
 * method also changes <code>yTicksCaptionStyle</code> to
 * <code>OBJECT</code>.
 */
public final void setYticksCaptionObject(ITickCaption object)
{
  yTicksCaptionObject = object;
  yTicksCaptionStyle = OBJECT;
}

/**
 * return the caption object for the y-axis. Undefined
 * behaviour, if <code>yTicksCaptionStyle != OBJECT</code>.
 */
public final ITickCaption getYticksCaptionObject()
{
  return yTicksCaptionObject;
}

/**
 * set, how the x-axis behaves when we attempt to plot a value
 * which is larger or smaller than xmax resp. xmin.
 *
 * @param behaviour: NONE, FIT, SCROLL are excepted
 * @exception IllegalArgumentException otherwise
 */
public final void setXaxisBehaviour(int behaviour)
  throws IllegalArgumentException
{
  if(behaviour >= NONE && behaviour <= SCROLL)
    xAxisBehaviour = behaviour;
  else
    throw new IllegalArgumentException();
  repaint();
}

/**
 * get , how the x-axis behaves when we attempt to plot a value
 */
public final int getXaxisBehaviour()
{
  return xAxisBehaviour;
}

/**
 * set, how the y-axis behaves when we attempt to plot a value
 * which is larger or smaller than ymax resp. ymin.
 *
 * @param behaviour: NONE, FIT are excepted
 * @exception IllegalArgumentException otherwise
 */
public final void setYaxisBehaviour(int behaviour)
{
  // scrolling y-axis isn't supported. Does it make sense anyway?
  if(behaviour == NONE || behaviour == FIT)
    yAxisBehaviour = behaviour;
  else
    throw new IllegalArgumentException();
  repaint();
}

/**
 * get , how the y-axis behaves when we attempt to plot a value
 */
public final int getYaxisBehaviour()
{
  return yAxisBehaviour;
}

/**
 * map values in [xmin, xmax] to [0, currentWidth]
 */
private int translateX(double d)
{
  return (int)((d - xmin) * currentWidth / (xmax - xmin)) + getInsets().left;
}

/**
 * map values in [ymin, ymax] to [0, currentHeight]
 */
private int translateY(double d)
{
  return (int) ((ymax - d) / (ymax - ymin) * currentHeight) + getInsets().top;
}

/*
 * for implementing the Observer interface. 
 * gets called by an Observable object, which notifies it's Observers about 
 * a change.
 */ 
public void update(Observable o,
                   Object arg)
{
  if(arg != null)
    {
      UpdateDescription reason = (UpdateDescription) arg;
      
      if(reason.reason == UpdateDescription.ALL ||
         reason.reason == UpdateDescription.DATA ||
         reason.reason == UpdateDescription.GRAPH_COLOR)
        {
          repaint();
        }
    }
  else
    repaint();
}

/*
 * paint everything
 */
public void paintComponent(Graphics gc)
{
  //paint background
  //super.paintComponent(gc);
  
  // compute drawing boundaries
  Insets insets = getInsets();
  currentWidth  = getWidth() - insets.left - insets.right - 1;
  currentHeight = getHeight() - insets.top - insets.bottom - 1;

  // save the original clipping area
  Shape orgClipRect = gc.getClip();
  
  // set the clipping area so we dont draw outside the insets.
  gc.setClip(insets.left, insets.top, currentWidth + 1, currentHeight + 1);
  
  double new_xmin = xmin;
  double new_xmax = xmax;
  double new_ymin = ymin;
  double new_ymax = ymax;

  if(xAxisBehaviour != NONE || yAxisBehaviour != NONE)
    {
      for(Enumeration enum = plots.elements(); enum.hasMoreElements();)
        {
          IDataSeries ds = (IDataSeries) enum.nextElement();
          if(ds.getXMin() < xmin)
            {
              new_xmin = ds.getXMin();
              // round down
              if(((int) (new_xmin / xTicksDistance)) != 
                 (new_xmin / xTicksDistance))
                {
                  new_xmin = ((int) ((new_xmin -xTicksDistance) /
                                     xTicksDistance)) * xTicksDistance;
                }
            }
          if(ds.getYMin() < ymin)
            {
              new_ymin = ds.getYMin();
              if(((int) (new_ymin / yTicksDistance)) != 
                 (new_ymin / yTicksDistance))
                {
                  // round down
                  new_ymin = ((int) ((new_ymin -yTicksDistance) / 
                                     yTicksDistance)) * yTicksDistance;
                }
            }
          if(ds.getXMax() > xmax)
            {
              new_xmax = ds.getXMax();
              
              if(((int) (new_xmax / xTicksDistance)) != 
                 (new_xmax / xTicksDistance))
                {
                  // round up to the next tick
                  new_xmax = ((int)((new_xmax + xTicksDistance) / 
                                    xTicksDistance)) * xTicksDistance;
                }
            }
          if(ds.getYMax() > ymax)
            {
              new_ymax = ds.getYMax();
              if(((int) (new_ymax / yTicksDistance)) != 
                 (new_ymax / yTicksDistance))
                {
                  // round up to the next tick
                  new_ymax = ((int)((new_ymax + yTicksDistance) / 
                                    yTicksDistance)) * yTicksDistance;
                }
            }
        }
    }
  
  if(xAxisBehaviour == FIT)
    {
      xmin = new_xmin;
      xmax = new_xmax;
    }
  else if(xAxisBehaviour == SCROLL)
    {
      xmin += new_xmax - xmax;
      xmax = new_xmax;
    }
  
  if(yAxisBehaviour == FIT)
    {
      ymin = new_ymin;
      ymax = new_ymax;
    }

  // draw coordinate system
  Color orgColor = gc.getColor();
  Font  orgFont  = gc.getFont();

  drawXaxis(gc);
  drawYaxis(gc);

  drawPlots(gc);
  
  // restore saved values
  gc.setColor(orgColor);
  gc.setFont(orgFont);
  gc.setClip(orgClipRect);
}

/**
 * draw x-axis. respects <code>xTicksCaptionStyle</code>
 */ 
private void drawXaxis(Graphics gc)
{
  Insets insets = getInsets();

  // draw x-axis
  int x1 = insets.left;
  int y1 = translateY(0.0);
  int x2 = insets.left + currentWidth;
  int y2 = y1;
  
  gc.setColor(axisColor);
  gc.setFont(axisFont);
  FontMetrics fm = gc.getFontMetrics(axisFont);
  
  gc.drawLine(x1, y1, x2, y2);

  //draw x-ticks and caption (if xTicksCaptionStyle != NONE)
  y1 = translateY(0.0);
  y2 = y1 - 2;
      
  for(double x = xmin; x <= xmax; x += xTicksDistance)
    {
      int x_java = translateX(x);
      
      // draw ticks
      if(ticksAsDottedLines)
        {
          int yStop = currentHeight + insets.top;

          // its not possible to draw dotted lines with Java
          // 1.1!!! That's why, we have to dot a line ourself.
          for(int y = insets.top; y <= yStop; y += 4)
            {
              gc.drawLine(x_java, y, x_java, y);
            }
        }
      else
        gc.drawLine(x_java, y1, x_java, y2);

      // draw caption
      if(xTicksCaptionStyle != NONE)
        {
          String s = xTicksCaptionObject.markTick(x);
          
          if(x == xmax) // don't draw outside the screen
            gc.drawString(s, 
                          x_java - fm.stringWidth(s) - 2, 
                          y1 + fm.getHeight());
          else
            gc.drawString(s, 
                          x_java + 2, 
                          y1 + fm.getHeight());
        }
    }
}


/**
 * draw x-axis. respects <code>xTicksCaptionStyle</code>
 */ 
private void drawYaxis(Graphics gc)
{
  Insets insets = getInsets();

  // draw y-axis
  int x1 = translateX(0.0);
  int y1 = insets.top;
  int x2 = x1;
  int y2 = insets.top + currentHeight;

  gc.setColor(axisColor);
  gc.setFont(axisFont);
  FontMetrics fm = gc.getFontMetrics(axisFont);
  
  gc.drawLine(x1, y1, x2, y2);

  //draw y-ticks and caption
  x1 = translateX(0.0);
  x2 = x1 + 2;
  
  for(double y = ymin; y <= ymax; y += yTicksDistance)
    {
      int y_java = translateY(y);

      if(ticksAsDottedLines)
        {
          int xStop = currentWidth + insets.left;
          for(int x = insets.left; x <= xStop; x += 4)
            {
              gc.drawLine(x, y_java, x, y_java);
            }
        }
      else
        gc.drawLine(x1, y_java, x2, y_java);

      if(yTicksCaptionStyle != NONE)
        {
          String s      = yTicksCaptionObject.markTick(y);
          int    x_java = x1 - fm.stringWidth(s) - 2;

          // if 0 lies outside the actual drawing window, print
          // the caption for y-axis on the left side of the area.
          if(x_java <= xmin)
            {
              x_java = translateX(xmin) + 2;
              if(y == 0.0)
                s = "";
            }
                              
          if(y == ymin)
            {
              gc.drawString(s, x_java, y_java - 2);
            }
          else
            {
              gc.drawString(s, x_java, y_java + fm.getHeight());
            }
        }
    }
}

/**
 * plot the data of all registered plots. Respects {x|y}AxisBehaviour.
 */
private void drawPlots(Graphics gc)
{
  int numPlots = plots.size();
  
  for(int i = 0; i < numPlots; i++)
    {
      IDataSeries plot     = (IDataSeries) plots.elementAt(i);
      Vector      data     = plot.getData();
      int         plotSize = data.size();

      // if the plot contains data draw it
      if(plotSize > 0)
        {
          DataSet lastDS    = (DataSet) data.lastElement();

          Color defaultColor = plot.getColor();
      
          for(int j = plotSize - 2; j >= 0; j--)
            {
              DataSet ds = (DataSet) data.elementAt(j);
              
              Color dsColor = ds.getColor();
              
              if(dsColor != null)
                gc.setColor(dsColor);
              else
                gc.setColor(defaultColor);
              
              if(ds.getX() < xmin)
                {
                  break;
                }
          
              gc.drawLine(translateX(lastDS.getX()), 
                          translateY(lastDS.getY()), 
                          translateX(ds.getX()), 
                          translateY(ds.getY()));
              lastDS = ds;
            }
        }
      
    }
}

/*
 * react on a l&f change
 */
public void updateUI()
{
  axisColor = UIManager.getColor("Label.foreground");
  axisFont  = 
    new Font(UIManager.getFont("Label.font").getName(), 
             UIManager.getFont("Label.font").getStyle(), 10);
  
  //repaint();
  
  super.updateUI();
}

/*
 * sizing stuff
 */
public Dimension getMinimumSize()
{
  return new Dimension(currentWidth, currentHeight);
}

public Dimension getPreferredSize()
{
  return getMinimumSize();
}

public Dimension getSize()
{
  return getMinimumSize();
}

public void setSize(Dimension d)
{
  setSize(d.width, d.height);
}

public void setSize(int width, int height)
{
  currentWidth  = width;
  currentHeight = height;

  super.setSize(width, height);
}

public void setBounds(Rectangle r)
{
  currentWidth = r.width;
  currentHeight = r.height;
  
  super.setBounds(r);
}
};
